package org.openxma.dsl.generator.helper;

import org.eclipse.emf.ecore.EObject;
import org.openxma.dsl.common.Strings;
import org.openxma.dsl.pom.model.Component;
import org.openxma.xmadsl.GuidesignExtension;

import at.spardat.xma.guidesign.BDAttribute;
import at.spardat.xma.guidesign.BDCollection;
import at.spardat.xma.guidesign.BusinessData;
import at.spardat.xma.guidesign.IBDAttachable;
import at.spardat.xma.guidesign.ValidInState;
import at.spardat.xma.guidesign.XMAComponent;
import at.spardat.xma.guidesign.XMAComposite;
import at.spardat.xma.guidesign.XMAFormData;
import at.spardat.xma.guidesign.XMAPage;
import at.spardat.xma.guidesign.flex.CollapseChain;
import at.spardat.xma.guidesign.flex.CompoundValidator;
import at.spardat.xma.guidesign.flex.Expression;
import at.spardat.xma.guidesign.flex.IGeneratable;
import at.spardat.xma.guidesign.flex.XMAApplicationContext;
import at.spardat.xma.guidesign.flex.XMAFunction;
import at.spardat.xma.guidesign.flex.XMAStateFlag;

public class CleanXMAComponentExtension {
	
	/**
	 * Remove all elements of a XMAComponent which are created from Xmadsl and links to them.
	 * 
	 * @param xmaComponent
	 */
	public static void removeElementsWhichAreCreatedFromXmadsl(XMAComponent xmaComponent, Component component) {		
		try {
			if (xmaComponent == null) {
				String modelName = "UNDEFINED";
				if (component != null) {
					modelName = component.getName();
				}
				throw new IllegalArgumentException("The parameter xmaComponent must not be null. Most probably the presentation model '"+modelName+"'is not valid.");
			}
			// Remove all generated Elements
			for (int i=xmaComponent.eContents().size()-1;i>=0;i--) {
				EObject child = xmaComponent.eContents().get(i);
				removeGeneratedObjects(child);
			}												
			// Remove empty BDCollection from XMAComponent
			if (xmaComponent instanceof BDCollection) {
				BDCollection bdc = (BDCollection)xmaComponent;
				if (bdc != null && bdc.getBusinessdata() == null) {
					Util.removeObject(xmaComponent);
				}
				return;
			}					
		} catch (Exception ex) {
			ex.printStackTrace();
			throw new RuntimeException("Can not remove generated elements",ex);			
		}					
	}
	
	/**
	 * Rekursively remove all generated elements which re created from Xmadsl and links to them
	 * 
	 * @param obj
	 */
	public static void removeGeneratedObjects(EObject obj) {
		
		// TODO introduce generated flag on page and refacture this code 
		if (obj instanceof XMAPage) {
			XMAPage page = (XMAPage)obj;			
			String namClass = page.getNamClass();
			String namInstance = page.getNamInstance();
			if ((Strings.uncapitalize(namClass)+GuidesignExtension.GENERATED_PAGE_EXTENSION).equals(namInstance)) {
				removeObject(obj);
			}
		}
		
		if (obj instanceof IGeneratable && ((IGeneratable)obj).isYnGenerated()) {
			removeObject(obj);
		} else if (obj instanceof CompoundValidator) {
			removeObject(obj);
		} else if (obj instanceof CollapseChain) {
			removeObject(obj);			
		} else if (obj instanceof ValidInState && ((ValidInState)obj).getValidator()==null) {
			removeObject(obj);
		} else if (obj instanceof XMAApplicationContext) {
			removeObject(obj);
		} else if (obj instanceof XMAFunction) {
			removeObject(obj);
		} else if (obj instanceof XMAStateFlag) {
			removeObject(obj);
		} else if (obj instanceof Expression) {
			removeObject(obj);
		} else if (obj instanceof XMAComposite && ((XMAComposite)obj).isYnSimpleLayout()) {
			XMAComposite xmaComposite = (XMAComposite)obj;
			for (int i=xmaComposite.eContents().size()-1;i>=0;i--) {				
				EObject child = obj.eContents().get(i);				
				if (child instanceof XMAFormData == false) {
					removeObject(child);
				}
			}
		} else if (obj instanceof IBDAttachable) {
			// Set the DataAttribute of not generated IBDAttachables to null
			IBDAttachable bDAttachable = (IBDAttachable)obj;
			BDAttribute attr = (BDAttribute)bDAttachable.getDataAttribute();
			if (attr != null) {
				BusinessData bd = (BusinessData)attr.eContainer();
				if (bd != null && bd.isYnGenerated()) {
					bDAttachable.setDataAttribute(null);
				}
			}			
		} 				
		for (int i=obj.eContents().size()-1;i>=0;i--) {
			EObject child = obj.eContents().get(i);
			removeGeneratedObjects(child);
		}
	}
	
	/**
	 * Remove the element
	 * 
	 * @param obj
	 */
	public static void removeObject(EObject obj) {
		Util.removeObject(obj);
	}

}
