package org.openxma.dsl.generator.helper;

import java.util.List;

import org.eclipse.emf.common.util.EList;
import org.eclipse.emf.ecore.EObject;
import org.eclipse.emf.ecore.util.EcoreUtil;
import org.openxma.dsl.core.model.SimpleType;
import org.openxma.dsl.dom.model.Attribute;
import org.openxma.dsl.pom.PomFactory;
import org.openxma.dsl.pom.model.AttachmentOwnPosition;
import org.openxma.dsl.pom.model.AttachmentProperty;
import org.openxma.dsl.pom.model.AttachmentSpecification;
import org.openxma.dsl.pom.model.Button;
import org.openxma.dsl.pom.model.CheckBox;
import org.openxma.dsl.pom.model.CombinedModelElement;
import org.openxma.dsl.pom.model.Combo;
import org.openxma.dsl.pom.model.ComposeData;
import org.openxma.dsl.pom.model.Content;
import org.openxma.dsl.pom.model.CustomizeableField;
import org.openxma.dsl.pom.model.FieldPartSpecification;
import org.openxma.dsl.pom.model.GuiElement;
import org.openxma.dsl.pom.model.HeightProperty;
import org.openxma.dsl.pom.model.IElementWithLayoutData;
import org.openxma.dsl.pom.model.IncludePanel;
import org.openxma.dsl.pom.model.Label;
import org.openxma.dsl.pom.model.LayoutDataProperty;
import org.openxma.dsl.pom.model.ListBox;
import org.openxma.dsl.pom.model.ModelElementSpecification;
import org.openxma.dsl.pom.model.Page;
import org.openxma.dsl.pom.model.RadioButton;
import org.openxma.dsl.pom.model.SelectedModelElement;
import org.openxma.dsl.pom.model.SetOfGuiElements;
import org.openxma.dsl.pom.model.SplitPanel;
import org.openxma.dsl.pom.model.StylePropertyBottom;
import org.openxma.dsl.pom.model.StylePropertyFieldPart;
import org.openxma.dsl.pom.model.StylePropertyHeight;
import org.openxma.dsl.pom.model.StylePropertyLeft;
import org.openxma.dsl.pom.model.StylePropertyRight;
import org.openxma.dsl.pom.model.StylePropertyTabulator;
import org.openxma.dsl.pom.model.StylePropertyTop;
import org.openxma.dsl.pom.model.StylePropertyWidth;
import org.openxma.dsl.pom.model.StyleSelectorType;
import org.openxma.dsl.pom.model.StyleSpecification;
import org.openxma.dsl.pom.model.StyleSpecificationProperty;
import org.openxma.dsl.pom.model.TabFolder;
import org.openxma.dsl.pom.model.TabPage;
import org.openxma.dsl.pom.model.Table;
import org.openxma.dsl.pom.model.TableColumn;
import org.openxma.dsl.pom.model.Text;
import org.openxma.dsl.pom.model.Tree;
import org.openxma.dsl.pom.model.WidthProperty;
import org.openxma.dsl.pom.model.XmadslComposite;
import org.openxma.dsl.pom.naming.PomNameProvider;
import org.openxma.dsl.pom.validation.PomStringConstants;


public class StylePropertyExtension {
	
	public static void setFieldPartSpecificationFromStyle(StyleSpecificationProperty sp, CustomizeableField cf) {
		if (cf.getParts().size() == 0) {
			StylePropertyFieldPart fp = (StylePropertyFieldPart)sp;						
			for (FieldPartSpecification fps : fp.getParts()) {							
				cf.getParts().add((FieldPartSpecification)EcoreUtil.copy(fps));
			}
		}
	}	
	

	public static void setLayoutPropertyFromStyle(StyleSpecificationProperty sp, IElementWithLayoutData eld) {		
		EList<LayoutDataProperty> lpl = eld.getLayoutProperties();					
		if (sp instanceof StylePropertyRight) {				
			setRightPropertyFromStyle(lpl, (StylePropertyRight)sp);
		} else if (sp instanceof StylePropertyLeft) {				
			setLeftPropertyFromStyle(lpl, (StylePropertyLeft)sp);
		} else if (sp instanceof StylePropertyTop) {				
			setTopPropertyFromStyle(lpl, (StylePropertyTop)sp);
		} else if (sp instanceof StylePropertyBottom) {				
			setBottomPropertyFromStyle(lpl, (StylePropertyBottom)sp);
		} else if (sp instanceof StylePropertyWidth) {				
			setWidthPropertyFromStyle(lpl, (StylePropertyWidth)sp, eld);
		} else if (sp instanceof StylePropertyHeight) {				
			setHeightPropertyFromStyle(lpl, (StylePropertyHeight)sp,eld);
		}
	}	
	
	private static void setRightPropertyFromStyle(EList<LayoutDataProperty> lpl, StylePropertyRight spr) {
		for (LayoutDataProperty ldp : lpl) {
			if (ldp instanceof AttachmentProperty && ((AttachmentProperty)ldp).getOwnPosition().getValue() == AttachmentOwnPosition.RIGHT_VALUE) {
				return;
			}
		}
		AttachmentProperty ap = PomFactory.eINSTANCE.createAttachmentProperty();
		ap.setOwnPosition(AttachmentOwnPosition.RIGHT);
		ap.setAttachmentSpecification((AttachmentSpecification)EcoreUtil.copy(spr.getAttachmentSpecification()));
		lpl.add(ap);
	}
	private static void setLeftPropertyFromStyle(EList<LayoutDataProperty> lpl, StylePropertyLeft spr) {
		for (LayoutDataProperty ldp : lpl) {
			if (ldp instanceof AttachmentProperty && ((AttachmentProperty)ldp).getOwnPosition().getValue() == AttachmentOwnPosition.LEFT_VALUE) {
				return;
			}
		}
		AttachmentProperty ap = PomFactory.eINSTANCE.createAttachmentProperty();
		ap.setOwnPosition(AttachmentOwnPosition.LEFT);
		ap.setAttachmentSpecification((AttachmentSpecification)EcoreUtil.copy(spr.getAttachmentSpecification()));
		lpl.add(ap);
	}
	
	private static void setTopPropertyFromStyle(EList<LayoutDataProperty> lpl, StylePropertyTop spr) {
		for (LayoutDataProperty ldp : lpl) {
			if (ldp instanceof AttachmentProperty && ((AttachmentProperty)ldp).getOwnPosition().getValue() == AttachmentOwnPosition.TOP_VALUE) {
				return;
			}
		}
		AttachmentProperty ap = PomFactory.eINSTANCE.createAttachmentProperty();
		ap.setOwnPosition(AttachmentOwnPosition.TOP);
		ap.setAttachmentSpecification((AttachmentSpecification)EcoreUtil.copy(spr.getAttachmentSpecification()));
		lpl.add(ap);
	}	
	
	private static void setBottomPropertyFromStyle(EList<LayoutDataProperty> lpl, StylePropertyBottom spr) {
		for (LayoutDataProperty ldp : lpl) {
			if (ldp instanceof AttachmentProperty && ((AttachmentProperty)ldp).getOwnPosition().getValue() == AttachmentOwnPosition.BOTTOM_VALUE) {
				return;
			}
		}
		AttachmentProperty ap = PomFactory.eINSTANCE.createAttachmentProperty();
		ap.setOwnPosition(AttachmentOwnPosition.BOTTOM);
		ap.setAttachmentSpecification((AttachmentSpecification)EcoreUtil.copy(spr.getAttachmentSpecification()));
		lpl.add(ap);
	}

	private static void setWidthPropertyFromStyle(EList<LayoutDataProperty> lpl, StylePropertyWidth spr, IElementWithLayoutData eld) {
		for (LayoutDataProperty ldp : lpl) {
			if (ldp instanceof WidthProperty) {
				if (((WidthProperty)ldp).getWidth()>0 && LayoutDataExtension.widthCanBeOverWrittenByWidthFromStyle(eld)==false) return;
				lpl.remove(ldp);				
				break;
			}
		}
		WidthProperty ap = PomFactory.eINSTANCE.createWidthProperty();
		ap.setWidth(spr.getWidth());
		LayoutDataExtension.setWidthIsFromStyle(eld);   		
		ap.setWidthInChar(spr.getWidthInChar());
		ap.setNegative(spr.isNegativeAfterChar());
		lpl.add(ap);
	}
	
	private static void setHeightPropertyFromStyle(EList<LayoutDataProperty> lpl, StylePropertyHeight spr, IElementWithLayoutData eld) {
		for (LayoutDataProperty ldp : lpl) {
			if (ldp instanceof HeightProperty) {	
			    if (((HeightProperty)ldp).getHeight()>0 && LayoutDataExtension.heightCanBeOverWrittenByHeightFromStyle(eld)==false) return;			    
				lpl.remove(ldp);				
				break;
			}
		}
		HeightProperty ap = PomFactory.eINSTANCE.createHeightProperty();
		ap.setHeight(spr.getHeight());
		LayoutDataExtension.setHeightIsFromStyle(eld);					
		lpl.add(ap);		
	}

	public static boolean fitsSelector(EObject obj,ModelElementSpecification mes) {
		int ci = mes.getCombinedElements().size()-1;
		CombinedModelElement cme = null;			
		if (ci >= 0) cme = mes.getCombinedElements().get(ci);			
		SelectedModelElement selectedModelElement = null;
		if (cme != null) {
			selectedModelElement = cme.getModelElement();
		} else {			
			 selectedModelElement = mes.getModelElement();				
		}
		if (matchesModelElement(obj, selectedModelElement)) {
			if (ci>=0) {
				return matchLeftElement(obj,mes,ci-1,cme);
			}
			return true;
		}		
		return false;
	}
	
	public static boolean fitsStyleSpecification(EObject obj, StyleSpecification styleSpecification) {
		if (obj == null) {
			return false;
		}
		List<ModelElementSpecification> mesList = styleSpecification.getModelElements();		
		if (mesList == null || mesList.size() == 0) return true;			
		for (ModelElementSpecification mes : mesList) {
			if (StylePropertyExtension.fitsSelector(obj,mes)) {
				return true;
			}
		}
		return false;
	}	
	
	private static boolean matchLeftElement(EObject obj, ModelElementSpecification mes, int ci, CombinedModelElement prevCme) {
		CombinedModelElement cme = null;			
		if (ci >= 0) cme = mes.getCombinedElements().get(ci);		
		SelectedModelElement selectedModelElement = null;
		if (cme != null) {
			selectedModelElement = cme.getModelElement();
		} else {			
			 selectedModelElement = mes.getModelElement();						
		}	
		int levelCount=0;
		EObject parent = obj.eContainer();
		while (parent != null) {
			int typeValue = typeValueOf(parent);
			if (typeValue >=0 ) {
				 if (prevCme != null && prevCme.isNext()) {
					 EObject previousChild = getPreviousChild(parent,obj);
					 if (matchesModelElement(previousChild,selectedModelElement)) {
							if (ci>=0) {
								return matchLeftElement(previousChild,mes,ci-1,cme);
							}
							return true;				 
					 }					 
					 return false;
				 }						
				levelCount++;			
				 if (matchesModelElement(parent,selectedModelElement)) {
						if (ci>=0) {
							return matchLeftElement(parent,mes,ci-1,cme);
						}
						return true;				 
				 }
				 if (prevCme != null && prevCme.isChild() && levelCount>0) {
					 return false;
				 }		 
			}
			if (parent.eContainer() != null) {
				parent = parent.eContainer();
			} else {
				return false;
			}
		}
		return false;
	}
	
	private static EObject getParentOfRelevantType(EObject obj) {
		EObject parent = obj.eContainer();
		while (parent != null) {
			int typeValue = typeValueOf(parent);
			if (typeValue >=0 ) {
				return parent;
			}
			parent = parent.eContainer();
		}	
		return null;		
	}
	
	private static EObject getRelativeChild(EObject parent, EObject obj, int delta) {
		if (parent instanceof XmadslComposite) {
			XmadslComposite comp = (XmadslComposite)parent;
			if (comp.getContent() != null) {
				int elementIndex = comp.getContent().getWidgetSetItems().indexOf(obj);
				if ((delta<0 && elementIndex > 0) || (delta>0 && elementIndex < comp.getContent().getWidgetSetItems().size()-1)) {
					return comp.getContent().getWidgetSetItems().get(elementIndex+delta);
				}
			}
			return null;
		}		
		if (parent instanceof SetOfGuiElements) {		
			SetOfGuiElements ge = (SetOfGuiElements)parent;
			int elementIndex = (ge.getElements().indexOf(obj));
			if ((delta<0 && elementIndex > 0) || (delta>0 && elementIndex < ge.getElements().size()-1)) {
				return ge.getElements().get(elementIndex+delta);
			}
			return null;
		}		
		return null;				
	}
	
	private static EObject getPreviousChild(EObject parent, EObject obj) {
		return getRelativeChild(parent,obj,-1);
	}
	
	private static EObject getNextChild(EObject parent, EObject obj) {
		return getRelativeChild(parent,obj,1);
	}
	
	private static EObject getObjectRelevantForName(EObject obj) {
	    if (obj instanceof ComposeData) {
	        return getObjectRelevantForName(obj.eContainer());
	    }
	    if (obj instanceof Content) {
	        return obj.eContainer();
	    }
	    return obj;
	}


	private static boolean matchesModelElement(EObject obj,	SelectedModelElement selectedModelElement) {
		if (selectedModelElement.getType() == StyleSelectorType.ANY || (fitType(obj,selectedModelElement) && fitDataType(obj,selectedModelElement))) {
			// Id selector
			if (selectedModelElement.getName() != null) {
				String selectedName = selectedModelElement.getName();
				EObject objectRelavantForName = getObjectRelevantForName(obj);
				String objName = PomNameProvider.getXmaName(objectRelavantForName);
				if (objName != null) {
					boolean prefix = selectedName.startsWith("*");
					boolean postfix = selectedName.endsWith("*");
					if (!prefix && !postfix) {
						return  selectedName.equals(objName);					
					}
					int beginIndex = (prefix)?1:0;
					int endIndex = (postfix)?selectedName.length()-1:selectedName.length();
					String pureName = selectedName.substring(beginIndex,endIndex);
					if (prefix && !postfix) {
						return objName.endsWith(pureName);
					}
					if (!prefix && postfix) {
						return objName.startsWith(pureName);
					}
					if (objName.indexOf(pureName) >= 0) {
						return true;
					}
				}
			// First child selector	
			} else if (selectedModelElement.getPseudoClass() != null) {
				if (PomStringConstants.FIRST_CHILD.equals(selectedModelElement.getPseudoClass())) {
					EObject parent = getParentOfRelevantType(obj);
					if (parent != null && getPreviousChild(parent,obj) == null) {
						return true;
					}
				// Last child selector	
				} else if (PomStringConstants.LAST_CHILD.equals(selectedModelElement.getPseudoClass())) {
					EObject parent = getParentOfRelevantType(obj);
					if (parent != null && getNextChild(parent,obj) == null) {
						return true;
					}
				// Middle child selector	
				} else if (PomStringConstants.MIDDLE_CHILD.equals(selectedModelElement.getPseudoClass())) {
					EObject parent = getParentOfRelevantType(obj);
					if (parent != null && getPreviousChild(parent,obj) != null && getNextChild(parent,obj) != null) {
						return true;
					}
				} else if (PomStringConstants.BROWSE.equals(selectedModelElement.getPseudoClass())) {
					// TODO: implement check for browse pseudo class
					return true;
				}
			// No additional selector to type
			} else {
				return true;
			}
		}
		return false;
	}
	
	private static int typeValueOf(EObject obj) {
		if (obj instanceof Label) return StyleSelectorType.LABEL_VALUE;
		if (obj instanceof CustomizeableField) return StyleSelectorType.FIELD_VALUE;		
		if (obj instanceof Text) return StyleSelectorType.TEXT_VALUE;		
		if (obj instanceof Combo) return StyleSelectorType.COMBO_VALUE;	
		if (obj instanceof Table) return StyleSelectorType.TABLE_VALUE;
		if (obj instanceof TableColumn) return StyleSelectorType.COLUMN_VALUE;		
		if (obj instanceof Tree) return StyleSelectorType.TREE_VALUE;		
		if (obj instanceof CheckBox) return StyleSelectorType.CHECKBOX_VALUE;
		if (obj instanceof RadioButton) return StyleSelectorType.RADIOBUTTON_VALUE;
		if (obj instanceof ListBox) return StyleSelectorType.LISTBOX_VALUE;
		if (obj instanceof SetOfGuiElements) return StyleSelectorType.CELL_VALUE;		
		if (obj instanceof XmadslComposite && isPanel((XmadslComposite)obj)) return StyleSelectorType.PANEL_VALUE;		
		if (obj instanceof XmadslComposite && isVPanel((XmadslComposite)obj)) return StyleSelectorType.VPANEL_VALUE;		
		if (obj instanceof XmadslComposite && isHPanel((XmadslComposite)obj)) return StyleSelectorType.HPANEL_VALUE;
		if (obj instanceof Button) return StyleSelectorType.BUTTON_VALUE;
		if (obj instanceof TabFolder) return StyleSelectorType.TABFOLDER_VALUE;
		if (obj instanceof TabPage) return StyleSelectorType.TABPAGE_VALUE;
		if (obj instanceof SplitPanel) return StyleSelectorType.SPLITPANEL_VALUE;
        if (obj instanceof IncludePanel) return StyleSelectorType.IPANEL_VALUE;		
		if (obj instanceof Page && (obj instanceof TabPage==false)) return StyleSelectorType.PAGE_VALUE;
		return -1;
	}
	
	private static boolean fitType(EObject obj, SelectedModelElement selectedModelElement) {
		int typeValue = selectedModelElement.getType().getValue();
		if (typeValue == StyleSelectorType.LABEL_VALUE && obj instanceof Label) return true;
		if (typeValue == StyleSelectorType.LABEL_VALUE && obj instanceof Label) return true;					
		if (typeValue == StyleSelectorType.FIELD_VALUE && obj instanceof CustomizeableField) return true;
		if (typeValue == StyleSelectorType.CELL_VALUE && obj instanceof SetOfGuiElements) return true;		
		if (typeValue == StyleSelectorType.TEXT_VALUE && obj instanceof Text) return true;			
		if (typeValue == StyleSelectorType.COMBO_VALUE && obj instanceof Combo) return true;			
		if (typeValue == StyleSelectorType.TABLE_VALUE && obj instanceof Table) return true;			
		if (typeValue == StyleSelectorType.COLUMN_VALUE && obj instanceof TableColumn) return true;		
		if (typeValue == StyleSelectorType.TREE_VALUE && obj instanceof Tree) return true;			
		if (typeValue == StyleSelectorType.CHECKBOX_VALUE && obj instanceof CheckBox) return true;			
		if (typeValue == StyleSelectorType.RADIOBUTTON_VALUE && obj instanceof RadioButton) return true;			
		if (typeValue == StyleSelectorType.LISTBOX_VALUE && obj instanceof ListBox) return true;			
		if (typeValue == StyleSelectorType.PANEL_VALUE && obj instanceof XmadslComposite && isPanel((XmadslComposite)obj)) return true;			
		if (typeValue == StyleSelectorType.VPANEL_VALUE && obj instanceof XmadslComposite && isVPanel((XmadslComposite)obj)) return true;			
		if (typeValue == StyleSelectorType.HPANEL_VALUE && obj instanceof XmadslComposite && isHPanel((XmadslComposite)obj)) return true;
		if (typeValue == StyleSelectorType.BUTTON_VALUE && obj instanceof Button) return true;		
		if (typeValue == StyleSelectorType.SPLITPANEL_VALUE && obj instanceof SplitPanel) return true;
		if (typeValue == StyleSelectorType.TABFOLDER_VALUE && obj instanceof TabFolder) return true;	
		if (typeValue == StyleSelectorType.TABPAGE_VALUE && obj instanceof TabPage) return true;
        if (typeValue == StyleSelectorType.IPANEL_VALUE && obj instanceof IncludePanel) return true;      		
		if (typeValue == StyleSelectorType.PAGE_VALUE && obj instanceof Page && (obj instanceof TabPage==false)) return true;
		// Container can not be selected itself, instead the containing element (panel, vpanel, hpanel or any kind of page
		// can be selected. Therefore for a Content object the container has to match.
		if (obj instanceof Content) {
		    EObject contentContainer = obj.eContainer();
		    return fitType(contentContainer,selectedModelElement);		    
		}
        // ComposeData can not be selected itself, instead a containing element can be selected. Currently the
		// containing element will be a Content which itself is handled in a special way.		
		if (obj instanceof ComposeData) {
		    EObject contentContainer = obj.eContainer();
            return fitType(contentContainer,selectedModelElement);  		    		    
		}
		return false;				
	}
	
	private static boolean fitDataType(EObject obj, SelectedModelElement selectedModelElement) {
		if (obj instanceof CustomizeableField && selectedModelElement.getSimpleType() != null) {
			CustomizeableField cf = (CustomizeableField)obj;
			Attribute attr = cf.getAttribute();
			if (attr != null && attr.getType() != null && attr.getType().getDataType() instanceof SimpleType) {
				SimpleType type = (SimpleType)attr.getType().getDataType();
				if (type.getName().equals(selectedModelElement.getSimpleType().getName())) {
					return true;
				} else {
					while (type.getTypeDefinition() != null && type.getTypeDefinition().getDefinition() != null) {
						type = (SimpleType)type.getTypeDefinition().getDefinition().getDataType();
						if (type.getName().equals(selectedModelElement.getSimpleType().getName())) {
							return true;
						}
					}
				}
			}			
			return false;
		}
		return true;
	}
	
	private static boolean isPanel(XmadslComposite xc) {
		if (xc.isHorizontalPartitioning()) return false;
		if (xc.isVerticalPartitioning()) return false;
		return true;
	}
	
	private static boolean isVPanel(XmadslComposite xc) {
		if (xc.isVerticalPartitioning()) return true;
		return false;
	}
	
	private static boolean isHPanel(XmadslComposite xc) {
		if (xc.isHorizontalPartitioning()) return true;
		return false;
	}	

	public static void setTabulatorFromStyle(StylePropertyTabulator sp,	GuiElement gel) {
		if (gel instanceof XmadslComposite) {
			XmadslComposite xc = (XmadslComposite)gel;
			if (xc.getContent() == null) {
				xc.setContent(PomFactory.eINSTANCE.createContent());
			}
			if (xc.getContent().getComposeLayout() == null) {
				xc.getContent().setComposeLayout(PomFactory.eINSTANCE.createComposeData());
			}
			if (xc.getContent().getComposeLayout().getTabulators().size() == 0) {
				xc.getContent().getComposeLayout().getTabulators().addAll(EcoreUtil.copyAll(sp.getTabulators()));
			}
		}		
	}			

}
